/*
   ldap_utils.c 
   general functions for LDAP
	 
   Author: Casper Pedersen
           cp@c-note.dk

*/

// ------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <lber.h>
#include <ldap_cdefs.h>
#include <ldap.h>
#include <sys/time.h>

// ------------------------------------------------------------------------
#include "defines.h"

// ------------------------------------------------------------------------
// Open LDAP connection, and do a bind.
// To get around Active Directory ACL problems. Annonymous bind's are
// normaly disabled on AD, as Microsoft are yet to figure out how to 
// use ACL's on public binds.
// ------------------------------------------------------------------------
LDAP* ldapOpenBind( struct config *conf)
{
	int rc;
	LDAP *ld;
	
	if( (ld = ldap_open( conf->ldapServer, conf->ldapPort )) == NULL )
	{
		logging( RUN, "- cannot open to %s:%s\n", conf->ldapServer, conf->ldapPort );
		exit(0);
	}
	
  logging( DEBUG, "- connected to ldapServer %s:%d", conf->ldapServer, conf->ldapPort);

	if( conf->is_bindDN && conf->is_bindPWD )
	{		
		rc = ldap_simple_bind_s( ld, conf->bindDN, conf->bindPWD );
		if( rc != LDAP_SUCCESS )
		{
			logging( RUN, "- cannot bind as %s - (%d) %s\n", conf->bindDN, rc, ldap_err2string(rc) );
			exit( 0 );
		}
		logging( DEBUG, "- bound to %s as %s\n", conf->ldapServer, conf->bindDN );
	}

	return ld;	
}

// ------------------------------------------------------------------------
// Open LDAP connection.
// ------------------------------------------------------------------------
LDAP* ldapOpen( struct config *conf)
{
	int rc;
	LDAP *ld;
	
	if( (ld = ldap_init( conf->ldapServer, conf->ldapPort )) == NULL )
	{
		logging( RUN, "- cannot open to %s:%s\n", conf->ldapServer, conf->ldapPort );
		exit(0);
	}
  logging( DEBUG, "- connected to ldapServer %s:%d", conf->ldapServer, conf->ldapPort);

	return ld;	
}


// ------------------------------------------------------------------------
// 1. Are we bound -> do a new connection
// 2. Otherwise use the current connection.
// ------------------------------------------------------------------------
int canBind(struct config *conf, LDAP *ld, char *dn, char *passwd)
{
  int rc; 
  
	LDAP  *lld = ld;
	int		local_open = 0;
	
	if( conf->is_bindDN && conf->is_bindPWD )
	{		
		local_open = 1;
		if( (lld = ldapOpen( conf )) == NULL )
			return 0;
	}

  logging( INFO, "- password check for %s", dn );
	rc = ldap_simple_bind_s( lld, dn, passwd );
  if( rc != LDAP_SUCCESS )
  {
    logging( DEBUG, "- password not OK!");
		
		if( local_open ) {
			ldap_unbind(lld);
		}	
			
    return rc;
  }
	
	if( local_open ) {
		ldap_unbind(lld);
	}	
	
  logging( DEBUG, "- password OK!");
  return 0;
}

// ------------------------------------------------------------------------
// Is group DN provided.
// ------------------------------------------------------------------------
int isDNGrp( char *grp )
{
  char **val;
  
  val = ldap_explode_dn( grp, 0 );
  
  if( strchr( val[0], '=') == NULL )
  {
    logging( DEBUG, "- no full group");
    return 0;
  }
  
  if( val[0][0] != '\0' )
  {
    int i = 0;
    while( val[i] != NULL )
    {
      logging( DEBUG, "- (%d) %s", i, val[i] );
      i++;
    }
    return 1;
  }
  
  free(val);

  return 0;
}

// ------------------------------------------------------------------------
// Find the DN of the group if not provided. 
// Crappy code, does not check if the group exist if DN is provided.
// ------------------------------------------------------------------------
int getGroupDN( LDAP *ld, struct config *conf, char* grpDN )
{
  LDAPMessage *searchResult, *entry;
  struct timeval timeOut;
  char searchStr[512];
  char *dn;
  int rc;
  int found = 0;
	
  timeOut.tv_sec = 10L;
  timeOut.tv_usec = 0L;

  if( isDNGrp( conf->pxyGroup ) )
  {
    logging( DEBUG, "- group has DN");
    strcpy( grpDN, conf->pxyGroup );
    return 1;
  }

  logging( DEBUG, "- group has not DN");

  sprintf(searchStr, "(&(objectClass=groupOfNames) (| (cn=%s)))", conf->pxyGroup);

  logging( DEBUG, "- search for: %s", searchStr );
  	
  rc = ldap_search_st( ld,
										   conf->searchBase,
											 LDAP_SCOPE_SUBTREE,
											 searchStr,
										   NULL, 0,
											 &timeOut,
											 &searchResult );

  logging( DEBUG, "- search done");

  if( ldap_count_entries( ld, searchResult ) == 0 )
  {
    return 0;
  }
	
  logging( DEBUG, "- found an entry");
  
  entry = ldap_first_entry( ld, searchResult );											 
  if( entry != NULL )
  {
    dn = ldap_get_dn( ld, entry );
    strcpy( grpDN, dn );
    logging( DEBUG, "- entry found: %s", grpDN );
    free( dn );
    found = 1;
  }
					
  ldap_msgfree(searchResult);
	
  return found;
}

// ------------------------------------------------------------------------
// Is CN an Alias 
// ------------------------------------------------------------------------
int isAlias( LDAP *ld, LDAPMessage *entry, char *dn )
{
  char **ret;
  ret = ldap_get_values( ld, entry, "aliasedObjectName" );
	
	if( ret != NULL )
  {	
		ldap_value_free( ret );
	  return 1;
	} 
  return 0;
}

// ------------------------------------------------------------------------
// Get the DN from the Alias object.
// ------------------------------------------------------------------------
int getDNfromAlias( LDAP *ld, LDAPMessage *entry, char *dn )
{
  BerElement **berptr;
	char **ret;
	ret = ldap_get_values( ld, entry, "aliasedObjectName" );
	if ( ret != NULL )
	{
	  strcpy( dn, ret[0] );
		ldap_value_free( ret );
		return 1;
	}
	
  return 0;
}

// ------------------------------------------------------------------------
// New code
// ------------------------------------------------------------------------
int getUserDN( LDAP *ld, struct config *conf, char* uid, char* udn )
{
  char searchStr[512];
  LDAPMessage *searchResult, *entry;
  struct timeval timeOut;
  char *dn;
  int rc;
  int found = 0;
	
  timeOut.tv_sec = 10L;
  timeOut.tv_usec = 0L;
	
  sprintf (searchStr, BLDUSRSRC, conf->userAttr, uid);
  logging( DEBUG, "- searchstr: %s", searchStr );
  logging( DEBUG, "- start searching for %s: %s", conf->userAttr, uid );
	
  rc = ldap_search_st (ld,
                       conf->searchBase,
		                   LDAP_SCOPE_SUBTREE,
		                   searchStr, 
		                   NULL, 
		                   0, 
		                   &timeOut, 
		                   &searchResult);
  	
  logging( DEBUG, "- search done");

  if( ldap_count_entries(ld, searchResult) == 0 )
  {
    logging( WARN, "- user \"%s\", not found!\n", uid);
    ldap_msgfree (searchResult); 
		return 0;
  }

  logging( DEBUG, "- found a entry");

	udn[0] = '\0';
	found = 0;
    
  if( (entry = ldap_first_entry(ld, searchResult )) != NULL )
  {
    if ( (dn = ldap_get_dn(ld,entry)) != NULL )		
    {

			if( isAlias( ld, entry, dn ) )
		  {
			  logging( RUN, "- DN is an alias" );
				if( !getDNfromAlias( ld, entry, udn ) )
				{
				  found = 0;
					logging( RUN, "- Attributte: \"aliasedObjectName\" not found" );
				}
				else
          logging( DEBUG, "- DN is %s", udn );		
		  }
			else
			{
        strcpy( udn, dn );
			}
				
      free(dn);
			
      logging( DEBUG, "- DN found: %s", udn );			
      found = 1;
    }	
  }

  ldap_msgfree (searchResult);    
	
  return found;	
}

// ------------------------------------------------------------------------
// New function, faster.
// Check if the user is member of group, instead of checking if user
// is in a group.
// ------------------------------------------------------------------------
int checkUserGroup( LDAP *ld, struct config *conf, char *gdn, char *dn )
{
  int rc;
	
  logging( DEBUG, "- is user %s in %s\n", dn, gdn );
  if( ldap_compare_s( ld, dn, conf->groupAttr, gdn ) == LDAP_COMPARE_TRUE )
  {
    logging( DEBUG, "- user \"%s\" is in Group \"%s\"", dn, gdn );
    return 1;		
  }
	
  logging( DEBUG, "- user \"%s\" is NOT in Group \"%s\"", dn, gdn );
  return 0;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
